require 'matrix'

def factorial(num)
  if num <= 0
    1
  else
    (1..num).to_a.inject(1){|product,n| product * n}
  end
end

def combinations(n,k)
  factorial(n).to_f / (factorial(k) * factorial(n - k))
end

def max(n,m)
  n > m ? n : m
end

def min(n,m)
  n < m ? n : m
end

#these next two bits (ha!) allow you to treat Boolean values like integers
#True is 1 and False is 0. It's comes in useful later
class TrueClass
  def to_i
    1
  end
end

class FalseClass
  def to_i
    0
  end
end

def dot(a, b)
  a.zip(b).inject(0){|sum,pair| sum + pair[0] * pair[1]}
end

class Edge
  attr_writer :weight
  attr_reader :from, :to, :directed, :weight
  def initialize(verts,directed=true,weight=1)
    @from = verts[0]
    @to = verts[1]
    @directed = directed
    @weight = weight
  end
  def connects(from,to)
    if @directed
      (@from == from) and (@to == to)
    else
      ((@from == from) and (@to == to)) or ((@from == to) and (@to == from))
    end
  end
  #does not take weight in to account on purpose
  def ==(other)
    (self.class == other.class) and (@directed == other.directed) and self.connects(other.from,other.to)
  end
  def make_undirected
    @directed = false
  end
  def other_end(vert)
    if vert == @from
      @to
    elsif vert == @to
      @from
    else
      nil
    end
  end
  def has_end(name)
    (@from == name) or (@to == name)
  end
end

class Vertex
  attr_reader :name, :edges
  def initialize(name)
    @name = name
    @edges = Array.new
  end
  #does not check to see if edge already exists
  def add_edge(edge)
    @edges.push(edge)
  end
  def delete_edge(edge)
    @edges.delete(edge)
  end
  def has_edge(edge)
    @edges.include?(edge)
  end
  def adj_to(name)
    @edges.find{|edge| edge.has_end(name)} != nil
  end
end

class Graph
public
  attr_reader :edges, :verticies
  #TODO: mechanism to initialize with weighted edges
  def initialize(edges=nil,directed=true)
    @directed = directed
    if edges != nil
      @verticies = Hash.new
      @edges = Array.new
      edges.flatten.uniq.each{|name| @verticies[name] = Vertex.new(name)}
#THIS PART IS UNTESTED!
      edges.each do |edge|
        temp = Edge.new(edge,directed)
        if @verticies[temp.from].has_edge(temp)
          #only need to update the weight once, because all edges all point to the same thing
          @verticies[temp.from].edges.find{|edge| edge == temp}.weight += temp.weight
        else
          @edges.push(temp)
          @verticies[temp.from].add_edge(temp)
          @verticies[temp.to].add_edge(temp)
        end
      end
    else
      @edges = Array.new
      @verticies = Hash.new
    end
  end
  def add_vertex(vertex)
    unless @verticies.include?(vertex)
      @verticies[vertex] = Vertex.new(vertex)
    end
    self
  end
  def add_edge(source, destination,weight=1)
    self.add_vertex(source)
    self.add_vertex(destination)
    temp = Edge.new([source, destination],@directed,weight)
    if @verticies[temp.from].has_edge(temp)
    #only need to update the weight once, because all edges all point to the same thing
      @verticies[temp.from].edges.find{|edge| edge == temp}.weight += temp.weight
    else
      @edges.push(temp)
      @verticies[temp.to].add_edge(temp)
      @verticies[temp.from].add_edge(temp)
    end
    self
  end
  def delete_edge(sorce, destination)
    temp = Edge.new([source, destination],@directed)
    @edges.delete(temp)
    @verticies[source].delete_edge(temp)
    @verticies[destination].delete_edge(temp)
    self
  end
  def delete_vertex(vertex)
    @verticies[vertex].edges.each{ |edge|
      @verticies[edge.other_end(vertex)].delete_edge(edge)
      @edges.delete(edge)}
    @verticies.delete(vertex)
    self
  end
  def v
    @verticies.size
  end
  def e
    @edges.size
  end
  def add_adj_edges(arr)
    first = arr.dup
    second = arr.dup
    first.pop
    second.shift
    first.zip(second).each{|edge| self.add_edge(edge[0],edge[1])}
  end
  def add_file(name)
    words = IO.read(name).split(%r{[,.:;!?]}).collect{|phrase| phrase.downcase.scan(/\w+/)}
    words.each{|arr| self.add_adj_edges(arr)}
  end
#this will have to be rewritten
#  def make_undirected
#    if @directed
#      @directed = false
#      @verticies.each_value do |vertex|
#        from = vertex.edges.find_all{|edge| edge.from == vertex.name}
#        to = vertex.edges.find_all{|edge| edge.from == to.name}
#        vertex.edges.each{|edge| edge.make_undirected}
#        from.each do |edge|
#          if to.include?(edge)
#            edge.weight += to.find{|e| e == edge}.weight
#            vertex.edges.delete(
#          end
#        end
#      end
#    end
#    self
#  end
  def num_to_name
    @verticies.keys.sort
  end
  def name_to_num
    lables = self.num_to_name
    ltable = Hash.new
    lables.each_index{|x| ltable[lables[x]] = x}
    ltable
  end
  def adj_matrix
puts "Starting to create adj matrix"
    ar = Array.new(@verticies.size){Array.new(@verticies.size,0.0)}
puts "array made (1/4)"
    ltable = self.name_to_num
puts "table made(2/4)"
    @edges.each do |edge|
      from = ltable[edge.from]
      to = ltable[edge.to]
      ar[from][to] += edge.weight
      unless @directed
        ar[to][from] = 1
      end
    end
puts "table populated (3/4)"
    t = Matrix.rows(ar,false)
puts "matrix made (4/4) Done creating adj matrix"
    t
  end
  def adj_list
    ltable = self.name_to_num
    @edges.collect{|edge| [ltable[edge.from],ltable[edge.to]]}
  end
  def hits
    adj = self.adj_matrix
#    adjt = adj.transpose
    h = Matrix.column_vector(Array.new(self.v,1.0 / Math.sqrt(self.v)))
    itr = 0
    begin
      itr +=1
      puts "Iteration #{itr}"
      hold = h.dup
puts "calculating a"
#      a = adjt * h
#no longer do it that way since, for large graphs, adjt is large and can take a long time to make
      a = Array.new(self.v,0.0)
      for i in 0..(self.v - 1) #i is the row of a
        for j in 0..(self.v - 1)
          a[i] += h[j,0] * adj[j,i]
        end
      end
      a = Matrix.column_vector(a)
puts "calculating h"
      h = adj * a
puts "normalizing a and h"
      #normalize the vectors... there must be a better way to do this...
      a = (1 / Math.sqrt((a.transpose * a)[0,0])) * a
      h = (1 / Math.sqrt((h.transpose * h)[0,0])) * h
      dot = (h.transpose * hold)[0,0]
      puts "The dot product of the current and former h vector is #{dot}"
    end while dot < 0.99
    return [a, h]
  end
  def hits_values
    ah = self.hits
    atable = Hash.new
    htable = Hash.new
    num_to_name.zip(ah[0].to_a.flatten).each{|arr| atable[arr[0]] = arr[1]}
    num_to_name.zip(ah[1].to_a.flatten).each{|arr| htable[arr[0]] = arr[1]}
    [atable,htable]
  end

end

class DocumentAnalyzer
#private
  def to_file(arr,filename)
    File.open(filename, "w") do |file|
      arr.each do |name, arr|
        file.print(name)
        arr.each{|val| file.print(",",val)}
        file.print("\n")
      end
    end
  end
  def avg_point(ranks)
    Matrix.column_vector(ranks.collect{|name,arr| arr.inject(0){|sum,value| sum + value}.to_f / arr.length})
  end
#public
  attr_reader :wordRanks, :documentNames, :documentGraphs, :cutwordRanks, :cutawordRanks
  def initialize
    @documentGraphs = Array.new
    @documentNames = Array.new
    @wordRanks = Hash.new {|hash,key| hash[key] = Array.new}
    @awordRanks = Hash.new {|hash,key| hash[key] = Array.new}
    @cutwordRanks = @wordRanks
    @cutawordRanks = @awordRanks
  end
  def add_document(name)
    @documentNames.push(name)
    temp = Graph.new
    puts "Making graph for #{name}..."
    temp.add_file(name)
    puts "The Graph has #{temp.v} verticies and #{temp.e} edges."
    @documentGraphs.push(temp)
    puts "Calculating hubs and authorities..."
    ah = temp.hits_values
    puts "Updating rankings..."
    ah[1].each do |label, value|
      arr = @wordRanks[label]
      #if 'label' isn't in @wordRanks, then we need to add zeros for the previous documents
      arr.push(0.0) while arr.length < (@documentNames.length - 1)
      arr.push(value)
    end
    #if a word wasn't in the document, add a zero
    @wordRanks.each_value{|arr| arr.push(0.0) unless arr.length == @documentNames.length}

    ah[0].each do |label, value|
      arr = @awordRanks[label]
      #if 'label' isn't in @wordRanks, then we need to add zeros for the previous documents
      arr.push(0.0) while arr.length < (@documentNames.length - 1)
      arr.push(value)
    end
    #if a word wasn't in the document, add a zero
    @awordRanks.each_value{|arr| arr.push(0.0) unless arr.length == @documentNames.length}

    self
  end
  def add_dir(dir_name)
    Dir.chdir(dir_name){Dir.foreach(dir_name){|file| self.add_document(file) if File.file?(file)}}
  end
  def cut_center_point
    self.avg_point(@cutwordRanks)
  end
  def center_point
    self.avg_point(@wordRanks)
  end
  def cut_scatter_matrix
    u = self.cut_center_point
    x = Matrix.rows(@cutwordRanks.collect{|name, arr| arr})
    a = Matrix.columns(x.column_vectors.collect{|cvect| (cvect - u).to_a.flatten})
    s = a * a.transpose
  end
  def cut_centered_points
    u = self.cut_center_point
    x = Matrix.rows(@cutwordRanks.collect{|name, arr| arr})
    a = Matrix.columns(x.column_vectors.collect{|cvect| (cvect - u).to_a.flatten}).column_vectors
  end
  def scatter_matrix
    u = self.center_point
    x = Matrix.rows(@wordRanks.collect{|name, arr| arr})
    a = Matrix.columns(x.column_vectors.collect{|cvect| (cvect - u).to_a.flatten})
    s = a * a.transpose
  end
  def hvector(index)
    @wordRanks.collect{|name,arr| arr[index]}
  end
  def avector(index)
    @awordRanks.collect{|name,arr| arr[index]}
  end
  def hfile_vector(file_name)
    index = @documentNames.index(file_name)
    self.hvector(index)
  end
  def afile_vector(file_name)
    index = @documentNames.index(file_name)
    self.avector(index)
  end
  def hnamevector
    @wordRanks.collect{|name,arr| name}
  end
  def anamevector
    @awordRanks.collect{|name,arr| name}
  end
  def cut(thresh)
    @cutwordRanks = Hash.new
    @cutawordRanks = Hash.new
    @wordRanks.each{|name,arr| @cutwordRanks[name] = arr if arr.inject(0){|sum,val| sum + val} > (thresh * arr.size)}
    @awordRanks.each{|name,arr| @cutawordRanks[name] = arr if arr.inject(0){|sum,val| sum + val} > (thresh * arr.size)}
  end
  def cutwordRanks_to_file(filename)
    self.to_file(@cutawordRanks,filename)
  end
  def wordRanks_to_file(filename)
    self.to_file(@wordRanks,filename)
  end
  def awordRanks_to_file(filename)
    self.to_file(@awordRanks,filename)
  end
  def sort_ranks_alphabetically
    @wordRanks = @wordRanks.sort{|a,b| a[0] <=> b[0]}
    @awordRanks = @awordRanks.sort{|a,b| a[0] <=> b[0]}
    @cutwordRanks = @cutwordRanks.sort{|a,b| a[0] <=> b[0]}
  end
  def sort_ranks_numerically
    @wordRanks = @wordRanks.sort{|a,b| b[1].inject(0){|sum,val| sum + val} <=> a[1].inject(0){|sum,val| sum + val}}
    @awordRanks = @awordRanks.sort{|a,b| b[1].inject(0){|sum,val| sum + val} <=> a[1].inject(0){|sum,val| sum + val}}
    @cutwordRanks = @cutwordRanks.sort{|a,b| b[1].inject(0){|sum,val| sum + val} <=> a[1].inject(0){|sum,val| sum + val}}
  end
  def sort_ranks_variance
    @wordRanks = @wordRanks.sort do |a,b|
      avgb = b[1].inject(0){|sum,val| sum + val}
      avga = a[1].inject(0){|sum,val| sum + val}
      varb = b[1].inject(0){|sum,val| sum + (val - avgb)**2}
      vara = a[1].inject(0){|sum,val| sum + (val - avga)**2}
      varb <=> vara
    end
    @awordRanks = @awordRanks.sort do |a,b|
      avgb = b[1].inject(0){|sum,val| sum + val}
      avga = a[1].inject(0){|sum,val| sum + val}
      varb = b[1].inject(0){|sum,val| sum + (val - avgb)**2}
      vara = a[1].inject(0){|sum,val| sum + (val - avga)**2}
      varb <=> vara
    end
  end
end
